unit Monitor;

{$IMPORTEDDATA OFF}

interface

var
  // Tokyo
  TimeLocationLatitude: Double = 35.6581;
  TimeLocationLongitude: Double = 139.7414;
  TimeLocationAltitude: Double = 0;

  UdpReceiverPort: Integer = 7531;

function GetValue(const Name: string): string;
function UseValue(const Name: string): string;
procedure Init;
procedure Update;

implementation

uses NSys, NLib, NDict, NDate, NSysStat, NNet, NForm, Solar;

type
  TUdpReceiver = class(TObject)
    Udp: TUdp;
    constructor Create;
    destructor Destroy; override;
    procedure OnRecv(const Param);
  end;

var
  Updated: Int64;
  UdpReceiver: TUdpReceiver;
  Values: TStrDict;
  SolarUpdatedDay: Integer;
  BootTimeCache: Int64;

function SIntDef(const S: string): Integer;
begin
  try
    Result := SInt(S);
  except
    on EValueError do Result := 0;
  end;
end;

function UseValue(const Name: string): string;
  procedure SplitName(const S: string; var Head, Tail: string);
  var
    i: Integer;
  begin
    for i := 1 to Length(S) do
      if S[i] = '.' then
      begin
        Head := Copy(S, 1, i-1);
        Tail := Copy(S, i+1, Length(S)-i);
        Exit;
      end;
    Head := '';
    Tail := S;
  end;
var
  p, h, t: string;
begin
  if not Values.HasKey(Name) then
    Values[Name] := '';
  t := Name;
  p := '';
  while True do
  begin
    SplitName(t, h, t);
    if h = '' then
      Break;
    p := p + h;
    if not Values.HasKey(p) then
      Values[p] := '';
    p := p + '.';
  end;
end;

function GetValue(const Name: string): string;
begin
  Result := Values[Name];
end;

procedure Init;
begin
  Values.Clear;
  Values['hostname'] := GetHostName;
  if Assigned(UdpReceiver) then
  begin
    UdpReceiver.Free;
    UdpReceiver := nil;
  end;
  Updated := 0;
  SolarUpdatedDay := 0;
  try
    BootTimeCache := BootTime;
  except
    on E: EOSError do ;
  end;
end;

procedure SetValue(const Name: string; Value: Integer); overload;
begin
  Values[Name] := IStr(Value);
end;

procedure SetValue(const Name: string; Value: Int64); overload;
begin
  Values[Name] := IStr(Value);
end;

procedure SetValue(const Name: string; const Value: string); overload;
begin
  Values[Name] := Value;
end;

var
  CpuOldTime: TCpuTimeArray;
procedure Cpu;
var
  i, cpucount: Integer;
  p: string;
  d, load, loadtotal: TCpuTime;
  now: TCpuTimeArray;
begin
  now := nil;
  try
    now := CpuTime;
  except
    on E: EOSError do Exit;
  end;
  if not Assigned(CpuOldTime) then
    SetLength(CpuOldTime, Length(now));
  loadtotal.Total := 0;
  loadtotal.User := 0;
  loadtotal.System := 0;
  loadtotal.Idle := 0;
  for i := 0 to High(now) do
  begin
    p := 'cpu.p' + IStr(i);
    d.Total := now[i].Total - CpuOldTime[i].Total;
    if d.Total > 0 then
    begin
      d.User := now[i].User - CpuOldTime[i].User;
      d.System := now[i].System - CpuOldTime[i].System;
      d.Idle := now[i].Idle - CpuOldTime[i].Idle;
      load.Total := Trunc(100 - d.Idle * 100 / d.Total);
      load.User := Trunc(d.User * 100 / d.Total);
      load.System := Trunc(d.System * 100 / d.Total);
      load.Idle := Trunc(d.Idle * 100 / d.Total);
      SetValue(p + '.load.total', load.Total);
      SetValue(p + '.load.user', load.User);
      SetValue(p + '.load.system', load.System);
      SetValue(p + '.load.idle', load.Idle);
      Inc(loadtotal.Total, load.total);
      Inc(loadtotal.User, load.user);
      Inc(loadtotal.System, load.system);
      Inc(loadtotal.Idle, load.idle);
    end;
    CpuOldTime[i] := now[i];
  end;
  cpucount := Length(now);
  SetValue('cpu.load.total', loadtotal.Total div cpucount);
  SetValue('cpu.load.user', loadtotal.User div cpucount);
  SetValue('cpu.load.system', loadtotal.System div cpucount);
  SetValue('cpu.load.idle', loadtotal.Idle div cpucount);
end;

var
  DiskFirstUpdate: Int64;
procedure Disk;
var
  d: Char;
  p: string;
begin
  for d := 'a' to 'z' do
  begin
    p := 'disk.' + d;
    if Values.HasKey(p) then
      try
        with StorageStat(d + ':/') do
        begin
          SetValue(p + '.total', Total);
          SetValue(p + '.used', Used);
          SetValue(p + '.free', Total - Used);
        end;
      except
        on E: EOSError do ;
      end;
  end;

  // delay due to DiskStats slow initialization
  if Updated = 0 then
  begin;
    DiskFirstUpdate := NanoSec + 1000000000;
    Exit;
  end;
  if DiskFirstUpdate > NanoSec then
    Exit;

  if Values.HasKey('disk.read') or Values.HasKey('disk.write') then
    with DiskStats do
    begin
      SetValue('disk.read.bytespersec', ReadBytesPerSec);
      SetValue('disk.write.bytespersec', WriteBytesPerSec);
    end;
end;

var
  HdTempUpdated: Int64;
procedure HdTemp;
var
  i, j: Integer;
  name: string;
  attrs: TSmartAttributeArray;
  t: Int64;
begin
  attrs := nil; // suppress warning
  t := NanoSec;
  if (Updated > 0) and ((t-HdTempUpdated)/1000000000 < 10*60) then
    Exit;
  HdTempUpdated := t;
  for i := 0 to 255 do
  begin
    name := 'hdtemp.hd' + IStr(i);
    if Values.HasKey(name) then
      try
        attrs := SmartAttributes(i);
        for j := 0 to High(attrs) do
          if attrs[j].Id = 194 then
          begin
            SetValue(name, attrs[j].Raw[0]);
            Exit;
          end;
      except
        on EOSError do ;
      end;
  end;
end;

procedure Mem;
begin
  try
    with MemStat do
    begin
      SetValue('mem.memtotal', MemTotal);
      SetValue('mem.memused', MemUsed);
      SetValue('mem.memfree', MemTotal - MemUsed);
      SetValue('mem.swaptotal', SwapTotal);
      SetValue('mem.swapused', SwapUsed);
      SetValue('mem.swapfree', SwapTotal - SwapUsed);
    end;
  except
    on E: EOSError do ;
  end;
end;

var
  NetOldInOctets, NetOldOutOctets: Cardinal;
procedure Net;
var
  i, j: Integer;
  ino, outo: Cardinal;
  t: Double;
  iftab: TIfTable;
  sameaddr: Boolean;
begin
  iftab := nil; // suppress warning
  if Values.HasKey('net.in') or Values.HasKey('net.out') then
  begin
    ino := 0;
    outo := 0;
    iftab := IfTable;
    for i := 0 to High(iftab) do
      with iftab[i] do
      begin
        // 6:Ethernet, 71:IEEE 802.11
        if (ifType <> 6) and (ifType <> 71) then
          Continue;

        if Int64(ifPhysAddress) shl 16 = 0 then
          Continue;

        sameaddr := False;
        for j := 0 to i - 1 do
        begin
          // 6:Ethernet, 71:IEEE 802.11
          if (iftab[j].ifType <> 6) and (iftab[j].ifType <> 71) then
            Continue;
          if Int64(iftab[j].ifPhysAddress) shl 16 =
              Int64(ifPhysAddress) shl 16 then
          begin
            sameaddr := True;
            break;
          end;
        end;
        if sameaddr then
          Continue;

        Inc(ino, ifInOctets);
        Inc(outo, ifOutOctets);
      end;

    //SetValue('net.in.bytes', ino);
    //SetValue('net.out.bytes', outo);
    if Updated > 0 then
    begin
      t := (NanoSec - Updated) / 1000000000;
      if t > 0 then
      begin
        SetValue('net.in.bytespersec', Trunc((ino - NetOldInOctets) / t));
        SetValue('net.out.bytespersec', Trunc((outo - NetOldOutOctets) / t));
      end;
    end;
    NetOldInOctets := ino;
    NetOldOutOctets := outo;
  end;

  if Values.HasKey('net.tcp') then
    with TcpStats do
    begin
      SetValue('net.tcp.rtoalgorithm', tcpRtoAlgorithm);
      SetValue('net.tcp.rtomin', tcpRtoMin);
      SetValue('net.tcp.rtomax', tcpRtoMax);
      SetValue('net.tcp.maxconn', tcpMaxConn);
      SetValue('net.tcp.activeopens', tcpActiveOpens);
      SetValue('net.tcp.passiveopens', tcpPassiveOpens);
      SetValue('net.tcp.attemptfails', tcpAttemptFails);
      SetValue('net.tcp.estabresets', tcpEstabResets);
      SetValue('net.tcp.currestab', tcpCurrEstab);
      SetValue('net.tcp.insegs', tcpInSegs);
      SetValue('net.tcp.outsegs', tcpOutSegs);
      SetValue('net.tcp.retranssegs', tcpRetransSegs);
      SetValue('net.tcp.inerrs', tcpInErrs);
      SetValue('net.tcp.outrsts', tcpOutRsts);
    end;
end;

procedure Power;
begin
  with PowerStat do
  begin
    case ACLine of
      aclsOffline: SetValue('power.source', 'Batt');
      aclsOnline:  SetValue('power.source', 'AC');
      aclsUnknown: SetValue('power.source', '?');
      aclsError:   SetValue('power.source', 'Err');
    end;
    case Battery of
      battCharging:    SetValue('power.battstat', 'charging');
      battNotCharging: SetValue('power.battstat', '');
      battNoBattery:   SetValue('power.battstat', 'no batt');
      battUnknown:     SetValue('power.battstat', '?');
    end;
    if BatteryLifePercent > -1 then
      SetValue('power.battlife.percent', BatteryLifePercent);
    if BatteryLifeTime > -1 then
    begin
      SetValue('power.battlife.hour', BatteryLifeTime div 3600);
      SetValue('power.battlife.min', BatteryLifeTime mod 3600 div 60);
      SetValue('power.battlife.sec', BatteryLifeTime mod 60);
    end
    else begin
      SetValue('power.battlife.hour', 0);
      SetValue('power.battlife.min', 0);
      SetValue('power.battlife.sec', 0);
    end;
  end;
end;

type
  TTimeZone = record
    Name: string;
    Bias: Integer;
  end;
var
  TimeZones: array of TTimeZone;
procedure Time;
  procedure setts(prefix: string; ts: TTime);
  const
    daynames: array [0..6] of string = (
      'Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'
    );
    monnames: array [1..12] of string = (
      'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
      'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'
    );
  var
    yday: Integer;
    d: TDate;
  begin
    with ts do
    begin
      SetValue(prefix + '.year', Year);
      SetValue(prefix + '.mon', Mon);
      SetValue(prefix + '.mday', MDay);
      SetValue(prefix + '.hour', Hour);
      SetValue(prefix + '.min', Min);
      SetValue(prefix + '.sec', Sec);
      SetValue(prefix + '.monname', monnames[Mon]);
      SetValue(prefix + '.dayname', daynames[WDay]);
      d := Date(Year, Mon, MDay);
      yday := d.YearDay;
      SetValue(prefix + '.yday', yday);
      SetValue(prefix + '.ydayremain',  365 + Ord(IsLeap(Year)) - yday);
      SetValue(prefix + '.shortyear', Year mod 1000 mod 100);
      SetValue(prefix + '.weeknum', d.ISOWeekNumber);
    end;
  end;
var
  t: Int64;
  key, h: string;
  b, cnt, i: Integer;
  rh, rm, sh, sm: Integer;
begin
  if Updated = 0 then // first call
  begin
    SetLength(TimeZones, 10);
    cnt := 0;
    b := 0; // suppress warning
    while Values.EachKeys(key) do
      if Length(key) = Length('time.p0000') then
      begin
        h := Copy(key, 1, 6);
        if h = 'time.p' then
          b := SIntDef(Copy(key, 7, 4))
        else if h = 'time.m' then
          b := -SIntDef(Copy(key, 7, 4))
        else
          Continue;
        if cnt = Length(TimeZones) then
          SetLength(TimeZones, cnt + cnt div 4);
        TimeZones[cnt].Name := key;
        TimeZones[cnt].Bias := b div 100 * 3600 + b mod 100 * 60;
        Inc(cnt);
      end;
    SetLength(TimeZones, cnt);
  end;

  t := Trunc(NanoSec/1000000000+0.5);
  setts('time', LocalTime(t));
  if Values.HasKey('time.utc') then
    setts('time.utc', UTCTime(t));
  for i := 0 to High(TimeZones) do
    setts(TimeZones[i].Name, UTCTime(t+TimeZones[i].Bias));

  if Values.HasKey('time.sunrise') or Values.HasKey('time.sunset') then
    with LocalTime(t) do
    begin
      if SolarUpdatedDay <> MDay then
      begin
        SolarCalculate(Year, Mon, MDay, UTCOffset/3600,
            TimeLocationLatitude, TimeLocationLongitude, TimeLocationAltitude,
            rh, rm, sh, sm);
        SolarUpdatedDay := MDay;
        SetValue('time.sunrise.hour', rh);
        SetValue('time.sunrise.min', rm);
        SetValue('time.sunset.hour', sh);
        SetValue('time.sunset.min', sm);
      end;
    end;
end;

constructor TUdpReceiver.Create;
begin
  Udp := TUdp.Create;
  Udp.Bind(UdpReceiverPort);
  Udp.OnRecv := OnRecv;
end;

destructor TUdpReceiver.Destroy;
begin
  Udp.Free;
end;

procedure TUdpReceiver.OnRecv(const Param);
var
  i: Integer;
  s, n, t, address: string;
begin
  SetString(s, nil, 1024);
  SetLength(s, Udp.Recv(s[1], Length(s), address));
  n := '';
  t := '';
  for i := 1 to Length(s) do
    if s[i] in [' ', #9] then
    begin
      if t = '' then
        Continue
      else if n = '' then
      begin
        n := t;
        t := '';
      end else
        t := t + s[i];
    end else
      t := t + s[i];
  SetValue('udp.'+n, t);
end;

procedure Uptime;
var
  t: Int64;
begin
  t := NSys.Time - BootTimeCache;
  SetValue('uptime.days', t div 86400);
  SetValue('uptime.hour', t mod 86400 div 3600);
  SetValue('uptime.min',  t mod 3600 div 60);
  SetValue('uptime.sec', t mod 60);
end;

procedure Update;
begin
  if Values.HasKey('cpu') then Cpu;
  if Values.HasKey('disk') then Disk;
  if Values.HasKey('mem') then Mem;
  if Values.HasKey('net') then Net;
  if Values.HasKey('power') then Power;
  if Values.HasKey('hdtemp') then HdTemp;
  if Values.HasKey('time') then Time;
  if Values.HasKey('udp') and not Assigned(UdpReceiver) then
    try
      UdpReceiver := TUdpReceiver.Create;
    except
      on E: EOSError do
      begin
        Values.Del('udp');
        raise;
      end;
    end;
  if Values.HasKey('uptime') then Uptime;
  Updated := NanoSec;
end;

initialization
  Values := TStrDict.Create;

finalization
  Values.Free;
  UdpReceiver.Free;

end.
