unit NSysStat;

{$IMPORTEDDATA OFF}
{$MINENUMSIZE 4}

interface

uses NTypes, NLib, NSys;

type
  TMemStat = record
    MemTotal: Int64;
    MemUsed: Int64;
    SwapTotal: Int64;
    SwapUsed: Int64;
  end;

  TStorageStat = record
    Path: WideString;
    Total: Int64;
    Used: Int64;
  end;

  TCpuTime = record
    Total: Int64;
    User: Int64;
    System: Int64;
    Idle: Int64;
  end;
  TCpuTimeArray = array of TCpuTime;

  TProcessStat = record
    Pid: TPid;
    PPid: TPid;
    Name: WideString;
    Memory: Int64;
    UserTime: Int64;
    SystemTime: Int64;
    Bits: Longint;
  end;
  TProcessStatArray = array of TProcessStat;

  TIfEntry = record
    ifIndex: Longint;
    ifDescr: array [0..255] of Char;
    ifType: Longint;
    ifMtu: Longint;
    ifSpeed: Cardinal;
    ifPhysAddress: array [0..7] of Byte;
    ifAdminStatus: Longint;
    ifOperStatus: Longint;
    ifLastChange: Cardinal;
    ifInOctets: Cardinal;
    ifInUcastPkts: Cardinal;
    ifInNUcastPkts: Cardinal;
    ifInDiscards: Cardinal;
    ifInErrors: Cardinal;
    ifInUnknownProtos: Cardinal;
    ifOutOctets: Cardinal;
    ifOutUcastPkts: Cardinal;
    ifOutNUcastPkts: Cardinal;
    ifOutDiscards: Cardinal;
    ifOutErrors: Cardinal;
    ifOutQLen: Cardinal;
  end;
  TIfTable = array of TIfEntry;

  TIpAddrEntry = record
    ipAdEntAddr: Cardinal;
    ipAdEntIfIndex: Longint;
    ipAdEntNetMask: Cardinal;
    ipAdEntBcastAddr: Longint;
    ipAdEntReasmMaxSize: Cardinal;
  end;
  TIpAddrTable = array of TIpAddrEntry;

  TTcpConnEntry = record
    tcpConnState: Longint;
    tcpConnLocalAddress: Cardinal;
    tcpConnLocalPort: Cardinal;
    tcpConnRemAddress: Cardinal;
    tcpConnRemPort: Cardinal;
    OwningPid: Cardinal;
  end;
  TTcpConnTable = array of TTcpConnEntry;

  TTcpStats = record
    tcpRtoAlgorithm: Longint;
    tcpRtoMin: Longint;
    tcpRtoMax: Longint;
    tcpMaxConn: Longint;
    tcpActiveOpens: Cardinal;
    tcpPassiveOpens: Cardinal;
    tcpAttemptFails: Cardinal;
    tcpEstabResets: Cardinal;
    tcpCurrEstab: Cardinal;
    tcpInSegs: Cardinal;
    tcpOutSegs: Cardinal;
    tcpRetransSegs: Cardinal;
    tcpInErrs: Cardinal;
    tcpOutRsts: Cardinal;
  end;

  TUdpEntry = record
    udpLocalAddress: Cardinal;
    udpLocalPort: Cardinal;
    OwningPid: Cardinal;
  end;
  TUdpTable = array of TUdpEntry;

  TUdpStats = record
    udpInDatagrams: Cardinal;
    udpNoPorts: Cardinal;
    udpInErrors: Cardinal;
    udpOutDatagrams: Cardinal;
  end;

  TDiskStats = record
    ReadBytesPerSec: Int64;
    WriteBytesPerSec: Int64;
  end;

  TACLineStat = (aclsOffline, aclsOnline, aclsUnknown, aclsError);
  TBatteryStat = (battCharging, battNotCharging, battNoBattery, battUnknown);
  TPowerStat = record
    ACLine: TACLineStat;
    Battery: TBatteryStat;
    BatteryLifePercent: Integer;
    BatteryLifeTime: Int64;
  end;

  TSmartAttribute = packed record
    Id: Byte;
    Flags: Word;
    Current: Byte;
    Worst: Byte;
    Raw: array [0..5] of Byte;
    Reserved: Byte;
  end;
  PSmartAttribute = ^TSmartAttribute;

  TSmartAttributeArray = array of TSmartAttribute;

function BootTime: Int64;
function MemStat: TMemStat;
function StorageStat(const Path: WideString): TStorageStat;
function ListStorage: TStringArray;
function CpuTime: TCpuTimeArray;
function ListProcessStat: TProcessStatArray;
function IfTable: TIfTable;
function IpAddrTable: TIpAddrTable;
function TcpConnTable(WithPid: Boolean = False): TTcpConnTable;
function TcpStats: TTcpStats;
function UdpTable(WithPid: Boolean = False): TUdpTable;
function UdpStats: TUdpStats;
function DiskStats: TDiskStats;
procedure DiskStatsStop;
function PowerStat: TPowerStat;
function SmartAttributes(Drive: Integer): TSmartAttributeArray;

implementation

uses Windows, TlHelp32, WinAdd, Ntdll, Psapi, Iphlpapi, PdhSub;

function BootTime: Int64;
var
  r: NTSTATUS;
  stodi: TSystemTimeOfDayInformation;
begin
  r := NtQuerySystemInformation(SystemTimeOfDayInformation, @stodi, SizeOf(stodi),
      nil);
  if r <> 0 then
    raise EOSError.Create(r);
  Result := (stodi.liKeBootTime - 116444736000000000) div 10000000;
end;

function MemStat: TMemStat;
var
  mex: TMemoryStatusEx;
begin
  with Result do begin
    mex.dwLength := SizeOf(mex);
    if not GlobalMemoryStatusEx(mex) then
      raise EOSError.Create;
    MemTotal := mex.ullTotalPhys;
    MemUsed := mex.ullTotalPhys - mex.ullAvailPhys;
    SwapTotal := mex.ullTotalPageFile - mex.ullTotalPhys;
    SwapUsed := SwapTotal - (mex.ullAvailPageFile - mex.ullAvailPhys);
    if SwapUsed < 0 then
      SwapUsed := 0;
  end;
end;

function StorageStat(const Path: WideString): TStorageStat;
var
  freebytes: TLargeInteger;
  olderrormode: UINT;
begin
  Result.Path := Path;
  olderrormode := SetErrorMode(SEM_FAILCRITICALERRORS);
  try
    with Result do begin
      if not GetDiskFreeSpaceExW(PWideChar(Path), nil, @Total, @freebytes) then
        raise EOSError.Create(Path);
      Used := Total - freebytes;
    end;
  finally
    SetErrorMode(olderrormode);
  end;
end;

function ListStorage: TStringArray;
var
  i: Integer;
  p: PChar;
  s: string;
begin
  SetLength(Result, 32);
  SetString(s, nil, GetLogicalDriveStrings(0, nil));
  GetLogicalDriveStrings(Length(s), PChar(s));
  p := PChar(s);
  i := 0;
  while (p^ <> #0) and (i < Length(Result)) do begin
    Result[i] := p;
    Inc(p, Length(p)+1);
    Inc(i);
  end;
  SetLength(Result, i);
end;

type
  PSystemProcessorTimes = ^TSystemProcessorTimes;
var
  CpuCount: Integer;
  CpuTimes: TCpuTimeArray;
  SystemProcessorTimesList: PSystemProcessorTimes;
function CpuTime: TCpuTimeArray;
var
  i: Integer;
  si: TSystemInfo;
  r: NTSTATUS;
  p: PSystemProcessorTimes;
begin
  if not Assigned(SystemProcessorTimesList) then begin
    GetSystemInfo(si);
    CpuCount := si.dwNumberOfProcessors;
    SetLength(CpuTimes, CpuCount);
    GetMem(SystemProcessorTimesList, SizeOf(TSystemProcessorTimes) * CpuCount);
  end;
  r := NtQuerySystemInformation(SystemProcessorTimes,
      SystemProcessorTimesList, SizeOf(TSystemProcessorTimes) * CpuCount, nil);
  if r <> 0 then
    raise EOSError.Create(r);
  p := SystemProcessorTimesList;
  for i := 0 to CpuCount - 1 do
    with CpuTimes[i] do begin
      Total := (p.liKernelIdleTime + p.liUserOtherTime) div 10000;
      User := p.liUserOtherTime div 10000;
      System := (p.liKernelIdleTime - p.liIdleTime) div 10000;
      Idle := p.liidleTime div 10000;
      Inc(p);
    end;
  Result := CpuTimes;
end;

function ListProcessStat: TProcessStatArray;
const
  PROCESSOR_ARCHITECTURE_AMD64 = 9;
  PROCESS_QUERY_LIMITED_INFORMATION = $1000;
var
  da: DWORD;
  i: Integer;
  p, s: THandle;
  pe: TProcessEntry32W;
  mc: TProcessMemoryCounters;
  ctime, etime, ktime, utime: TFileTime;
  osvi: TOSVersionInfo;
  token: THandle;
  luid: TLargeInteger;
  tp: TTokenPrivileges;
  len: DWORD;
  GetNativeSystemInfo: procedure (var SystemInfo: TSystemInfo); stdcall;
  IsWow64Process: function (hProcess: THandle; var Wow64Pricess: BOOL): BOOL;
      stdcall;
  wow64: BOOL;
  si: TSystemInfo;
begin
  GetNativeSystemInfo := GetProcAddress(GetModuleHandle('kernel32.dll'),
      'GetNativeSystemInfo');
  IsWow64Process := GetProcAddress(GetModuleHandle('kernel32.dll'),
      'IsWow64Process');

  FillChar(si, SizeOf(si), 0);
  if (Assigned(GetNativeSystemInfo)) then
    GetNativeSystemInfo(si);

  len := 0;
  if OpenProcessToken(GetCurrentProcess, TOKEN_ADJUST_PRIVILEGES, token) then begin
    LookupPrivilegeValue(nil, 'SeDebugPrivilege', luid);
    tp.PrivilegeCount := 1;
    tp.Privileges[0].Luid := luid;
    tp.Privileges[0].Attributes := SE_PRIVILEGE_ENABLED;
    AdjustTokenPrivileges(token, False, tp, 0, nil, len);
    CloseHandle(token);
  end;

  i := 0;
  SetLength(Result, 1024);
  osvi.dwOSVersionInfoSize := SizeOf(osvi);
  GetVersionEx(osvi);
  if osvi.dwMajorVersion >= 6 then  // Vista or later
    da := PROCESS_QUERY_LIMITED_INFORMATION
  else
    //da := PROCESS_QUERY_INFORMATION or PROCESS_VM_READ;
    da := PROCESS_QUERY_INFORMATION;

  s := CreateToolHelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  pe.dwSize := SizeOf(pe);
  Process32FirstW(s, pe);
  repeat
    if i = Length(Result) then
      SetLength(Result, Length(Result) + Length(Result) div 4);
    with Result[i] do begin
      Inc(i);
      Pid := pe.th32ProcessID;
      PPid := pe.th32ParentProcessID;
      Name := pe.szExeFile;
      p := OpenProcess(da, False, Pid);
      if p = 0 then
        Continue;
      if Assigned(GetProcessMemoryInfo) then
        if GetProcessMemoryInfo(p, @mc, SizeOf(mc)) then
          Memory := mc.WorkingSetSize;
      if GetProcessTimes(p, ctime, etime, ktime, utime) then begin
        UserTime := Int64(utime) div 10000000;
        SystemTime := Int64(ktime) div 10000000;
      end;
      if Assigned(IsWow64Process) then
        IsWow64Process(p, wow64)
      else
        wow64 := False;
      if (si.wProcessorArchitecture = PROCESSOR_ARCHITECTURE_AMD64) and
          not wow64 then
        Bits := 64
      else
        Bits := 32;
      CloseHandle(p);
    end;
  until not Process32NextW(s, pe);
  SetLength(Result, i);
end;

function IfTable: TIfTable;
var
  i: Integer;
  r, size: DWORD;
  tab: PMibIfTable;
begin
  size := SizeOf(tab^);
  GetMem(tab, size);
  try
    r := GetIfTable(tab, size, True);
    if r = ERROR_INSUFFICIENT_BUFFER then begin
      ReallocMem(tab, size);
      r := GetIfTable(tab, size, True);
    end;
    if r <> NO_ERROR then
      raise EOSError.Create;
    SetLength(Result, tab.dwNumEntries);
    for i := 0 to tab.dwNumEntries - 1 do
      with tab.table[i] do
      begin
        Result[i].ifIndex := dwIndex;
        Move(bDescr, Result[i].ifDescr, SizeOf(Result[i].ifDescr));
        Result[i].ifType := dwType;
        Result[i].ifMtu := dwMtu;
        Result[i].ifSpeed := dwSpeed; 
        Move(bPhysAddr, Result[i].ifPhysAddress,
            SizeOf(Result[i].ifPhysAddress));
        Result[i].ifAdminStatus := dwAdminStatus;
        Result[i].ifOperStatus := dwOperStatus;
        Result[i].ifLastChange := dwLastChange;
        Result[i].ifInOctets := dwInOctets;
        Result[i].ifInUcastPkts := dwInUcastPkts;
        Result[i].ifInNUcastPkts := dwInNUcastPkts;
        Result[i].ifInDiscards := dwInDiscards;
        Result[i].ifInErrors := dwInErrors;
        Result[i].ifInUnknownProtos := dwInUnknownProtos;
        Result[i].ifOutOctets := dwOutOctets;
        Result[i].ifOutUcastPkts := dwOutUcastPkts;
        Result[i].ifOutNUcastPkts := dwOutNUcastPkts;
        Result[i].ifOutDiscards := dwOutDiscards;
        Result[i].ifOutErrors := dwOutErrors;
        Result[i].ifOutQLen := dwOutQLen;
      end;
  finally
    FreeMem(tab);
  end;
end;

function IpAddrTable: TIpAddrTable;
var
  i: Integer;
  r, size: DWORD;
  tab: PMibIpAddrTable;
begin
  size := SizeOf(tab^);
  GetMem(tab, size);
  try
    r := GetIpAddrTable(tab, size, True);
    if r = ERROR_INSUFFICIENT_BUFFER then begin
      ReallocMem(tab, size);
      r := GetIpAddrTable(tab, size, True);
    end;
    if r <> NO_ERROR then
      raise EOSError.Create;
    SetLength(Result, tab.dwNumEntries);
    for i := 0 to tab.dwNumEntries - 1 do
      with tab.table[i] do begin
        Result[i].ipAdEntAddr := dwAddr;
        Result[i].ipAdEntIfIndex := dwIndex;
        Result[i].ipAdEntNetMask := dwMask;
        Result[i].ipAdEntBcastAddr := dwBCastAddr;
        Result[i].ipAdEntReasmMaxSize := dwReasmSize;
      end;
  finally
    FreeMem(tab);
  end;
end;

const
  AF_INET = 2;

function TcpConnTable(WithPid: Boolean = False): TTcpConnTable;
var
  i: Integer;
  r, size: DWORD;
  tab: Pointer;
begin
  // GetExtendedTcpTable is supported on XP SP2, Vista and later
  if not Assigned(GetExtendedTcpTable) then
    WithPid := False;
  tab := nil;
  size := 0;
  try
    if WithPid then begin
      r := GetExtendedTcpTable(
          nil, PDWORD(size), True, AF_INET, TCP_TABLE_OWNER_PID_ALL, 0);
      if r = ERROR_INSUFFICIENT_BUFFER then begin
        GetMem(tab, size);
        r := GetExtendedTcpTable(
            tab, PDWORD(size), True, AF_INET, TCP_TABLE_OWNER_PID_ALL, 0);
      end;
    end else begin
      r := GetTcpTable(nil, size, True);
      if r = ERROR_INSUFFICIENT_BUFFER then begin
        GetMem(tab, size);
        r := GetTcpTable(tab, size, True);
      end;
    end;
    if r <> NO_ERROR then
      raise EOSError.Create;

    SetLength(Result, PMibTcpTable(tab).dwNumEntries);
    for i := 0 to PMibTcpTable(tab).dwNumEntries - 1 do
      if WithPid then
        with PMibTcpTableOwnerPid(tab).table[i] do begin
          Result[i].tcpConnState := dwState;
          Result[i].tcpConnLocalAddress := dwLocalAddr;
          Result[i].tcpConnLocalPort := Be16toH(dwLocalPort);
          Result[i].tcpConnRemAddress := dwRemoteAddr;
          Result[i].tcpConnRemPort := Be16toH(dwRemotePort);
          Result[i].OwningPid := dwOwningPid;
        end
      else
        with PMibTcpTable(tab).table[i] do begin
          Result[i].tcpConnState := dwState;
          Result[i].tcpConnLocalAddress := dwLocalAddr;
          Result[i].tcpConnLocalPort := Be16toH(dwLocalPort);
          Result[i].tcpConnRemAddress := dwRemoteAddr;
          Result[i].tcpConnRemPort := Be16toH(dwRemotePort);
        end;
  finally
    FreeMem(tab);
  end;
end;

function TcpStats: TTcpStats;
var
  stats: TMibTcpStats;
begin
  GetTcpStatistics(stats);
  with stats do begin
    Result.tcpRtoAlgorithm := dwRtoAlgorithm;
    Result.tcpRtoMin := dwRtoMin;
    Result.tcpRtoMax := dwRtoMax;
    Result.tcpMaxConn :=dwMaxConn;
    Result.tcpActiveOpens := dwActiveOpens;
    Result.tcpPassiveOpens := dwPassiveOpens;
    Result.tcpAttemptFails := dwAttemptFails;
    Result.tcpEstabResets := dwEstabResets;
    Result.tcpCurrEstab := dwCurrEstab;
    Result.tcpInSegs := dwInSegs;
    Result.tcpOutSegs := dwOutSegs;
    Result.tcpRetransSegs := dwRetransSegs;
    Result.tcpInErrs := dwInErrs;
    Result.tcpOutRsts := dwOutRsts;
  end;
end;

function UdpTable(WithPid: Boolean = False): TUdpTable;
var
  i: Integer;
  r, size: DWORD;
  tab: PMibUdpTable;
begin
  // GetExtendedUdpTable is supported on XP SP2, Vista and later
  if not Assigned(GetExtendedUdpTable) then
    WithPid := False;
  tab := nil;
  size := 0;
  try
    if WithPid then begin
      r := GetExtendedUdpTable(
          nil, PDWORD(size), True, AF_INET, UDP_TABLE_OWNER_PID, 0);
      if r = ERROR_INSUFFICIENT_BUFFER then begin
        GetMem(tab, size);
        r := GetExtendedUdpTable(
            tab, PDWORD(size), True, AF_INET, UDP_TABLE_OWNER_PID, 0);
      end;
    end else begin
      r := GetUdpTable(nil, size, True);
      if r = ERROR_INSUFFICIENT_BUFFER then begin
        GetMem(tab, size);
        r := GetUdpTable(tab, size, True);
      end;
    end;
    if r <> NO_ERROR then
      raise EOSError.Create;
    SetLength(Result, PMibUdpTable(tab).dwNumEntries);
    for i := 0 to PMibUdpTable(tab).dwNumEntries - 1 do
      if WithPid then
        with PMibUdpTableOwnerPid(tab).table[i] do begin
          Result[i].udpLocalAddress := dwLocalAddr;
          Result[i].udpLocalPort := Be16toH(dwLocalPort);
          Result[i].OwningPid := dwOwningPid;
        end
      else
        with PMibUdpTable(tab).table[i] do begin
          Result[i].udpLocalAddress := dwLocalAddr;
          Result[i].udpLocalPort := Be16toH(dwLocalPort);
        end;
  finally
    FreeMem(tab);
  end;
end;

function UdpStats: TUdpStats;
var
  stats: TMibUdpStats;
begin
  GetUdpStatistics(stats);
  with stats do begin
    Result.udpInDatagrams := dwInDatagrams;
    Result.udpNoPorts := dwNoPorts;
    Result.udpInErrors := dwInErrors;
    Result.udpOutDatagrams := dwOutDatagrams;
  end;
end;


type
  TPdhCounters = record
    ReadBytesPerSec: HCOUNTER;
    WriteBytesPerSec: HCOUNTER;
  end;

var
  PdhQuery: HQUERY;
  PdhCounters: TPdhCounters;

procedure DiskStatsStop;
begin
  if PdhQuery <> 0 then begin
    if PdhCounters.ReadBytesPerSec <> 0 then
      PdhRemoveCounter(PdhCounters.ReadBytesPerSec);
    if PdhCounters.WriteBytesPerSec <> 0 then
      PdhRemoveCounter(PdhCounters.WriteBytesPerSec);
    PdhCloseQuery(PdhQuery);
  end;
  PdhQuery := 0;
  PdhCounters.ReadBytesPerSec := 0;
  PdhCounters.WriteBytesPerSec := 0;
end;

function DiskStats: TDiskStats;
var
  v: TPdhFmtCounterValue;
begin
  if PdhQuery = 0 then begin
    PdhOpenQuery(nil, 0, PdhQuery);
    PdhAddCounter(PdhQuery, '\PhysicalDisk(_Total)\Disk Read Bytes/sec', 0,
        PdhCounters.ReadBytesPerSec);
    PdhAddCounter(PdhQuery, '\PhysicalDisk(_Total)\Disk Write Bytes/sec', 0,
        PdhCounters.WriteBytesPerSec);
    PdhCollectQueryData(PdhQuery);
    Result.ReadBytesPerSec := 0;
    Result.WriteBytesPerSec := 0;
    Exit;
  end;

  PdhCollectQueryData(PdhQuery);
  with PdhCounters do begin
    PdhGetFormattedCounterValue(ReadBytesPerSec, PDH_FMT_LARGE, nil, v); 
    Result.ReadBytesPerSec := v.largeValue;
    PdhGetFormattedCounterValue(WriteBytesPerSec, PDH_FMT_LARGE, nil, v); 
    Result.WriteBytesPerSec := v.largeValue;
  end;
end;

function PowerStat: TPowerStat;
var
  s: TSystemPowerStatus;
begin
  with Result do begin
    ACLine := aclsError;
    Battery := battUnknown;
    BatteryLifePercent := -1;
    BatteryLifeTime := -1;
    if not GetSystemPowerStatus(s) then
      Exit;

    case s.ACLineStatus of
      0: ACLine := aclsOffline;
      1: ACLine := aclsOnline;
      255: ACLine := aclsUnknown;
    end;

    if s.BatteryFlag <= 7 then
      Battery := battNotCharging
    else if s.BatteryFlag = 255 then
      Battery := battUnknown
    else begin
      if (s.BatteryFlag and 128) = 128 then
        Battery := battNoBattery;
      if (s.BatteryFlag and 8) = 8 then
        Battery := battCharging;
    end;

    if s.BatteryLifePercent > 100 then
      BatteryLifePercent := -1
    else
      BatteryLifePercent := s.BatteryLifePercent;

    if s.BatteryLifeTime = $ffffffff then
      BatteryLifeTime := -1
    else
      BatteryLifeTime := s.BatteryLifeTime;
  end;
end;

const
  IOCTL_STORAGE_PREDICT_FAILURE = $2D1100;
  MaxAttributeCount = 30;

type
  TStoragePredictFailure = record
    PredictFailure: ULONG;
    VendorSpecific: array [0..511] of Byte;
  end;

function SmartAttributes(Drive: Integer): TSmartAttributeArray;
var
  h: THandle;
  retsize: DWORD;
  spf: TStoragePredictFailure;
  i, j: Integer;
  pattr: PSmartAttribute;
  path: string;
begin
  SetLength(Result, 0);

  path := '//./PhysicalDrive' + IStr(Drive);

  h := CreateFile(PChar(path), 0, FILE_SHARE_READ or FILE_SHARE_WRITE, nil,
      OPEN_EXISTING, 0, 0);
  if h = INVALID_HANDLE_VALUE then
    raise EOSError.Create(path);
  try
    if not DeviceIoControl(h, IOCTL_STORAGE_PREDICT_FAILURE, nil, 0, @spf,
        SizeOf(spf), retsize, nil) then
      raise EOSError.Create(path);
  finally
    CloseHandle(h);
  end;
  if spf.PredictFailure <> 0 then
    Exit;

  pattr := @spf.VendorSpecific[2];
  SetLength(Result, MaxAttributeCount);
  j := 0;
  for i := 0 to High(Result) do begin
    if pattr.id <> 0 then begin
      Result[j] := pattr^;
      Inc(j);
    end;
    Inc(pattr);
  end;
  SetLength(Result, j);
end;

end.
